import os
import tempfile

from zsuite.file_operations import debug_file_path


def test_debug_file_path(monkeypatch):
    monkeypatch.setenv("FILE_DEBUG", "True")
    with tempfile.NamedTemporaryFile() as temp_file:
        temp_file_name = temp_file.name
        result = debug_file_path(temp_file_name, force_print=True)

        assert result["file_path"] == temp_file_name
        assert result["file_exists"] is True
        assert result["file_readable"] is True
        assert result["file_writable"] is True
        assert result["file_executable"] in [True, False]  # windows doesn't have executable bit
        assert result["file_is_directory"] is False
        assert result["parent_dir_exists"] is True
        assert result["parent_dir_writable"] is True
        assert result["parent_dir_readable"] is True
        assert result["is_absolute"] is True  # Temp files usually have absolute paths
        assert result["absolute_path"] == os.path.abspath(temp_file_name)



import time

import pytest

from zsuite import CircuitBreaker
from zsuite.exceptions import CircuitBreakerTripped


def test_increment():
    cb = CircuitBreaker(max_events=3, time_window=1)
    for _ in range(3):
        cb.increment()
    assert cb.count() == 3
    with pytest.raises(CircuitBreakerTripped):
        cb.increment()


def test_reset():
    cb = CircuitBreaker(max_events=3, time_window=1)
    cb.increment()
    assert cb.count() == 1
    cb.reset()
    assert cb.count() == 0


def test_time_window():
    cb = CircuitBreaker(max_events=3, time_window=1)
    cb.increment()
    assert cb.count() == 1
    time.sleep(1.1)  # Sleep to exceed the time window
    assert cb.count() == 0


def test_custom_exception():
    custom_exc = ValueError
    cb = CircuitBreaker(max_events=1, time_window=1, custom_exception=custom_exc)
    with pytest.raises(custom_exc):
        cb.increment()
        cb.increment()


def test_initialization_fails():
    with pytest.raises(ValueError):
        CircuitBreaker(max_events=None, time_window=1)
    with pytest.raises(ValueError):
        CircuitBreaker(max_events=3, time_window=None)



from datetime import date, datetime, timezone
from unittest.mock import patch
from zoneinfo import ZoneInfo

import pytest

from zsuite.exceptions import UndeterminedBool
from zsuite.helpers import (
    fuzzy_bool,
    normalize_timestamp,
    now_utc,
)


def test_fuzzy_bool():
    cases = [
        (True, True),
        (False, False),
        ("True", True),
        ("False", False),
        ("T", True),
        ("F", False),
        ("", None),
        (1, True),
        (0, False),
        ("Yes", True),
        ("No", False),
        ("active", True),
        ("inactive", False),
        ("Enabled", True),
        ("Disabled", False),
    ]
    for testcase, expected in cases:
        assert fuzzy_bool(testcase) is expected

    badcases = []
    for x in badcases:
        with pytest.raises(UndeterminedBool):
            fuzzy_bool(x)


def test_now_utc_not_naive():
    with patch("zsuite.helpers.datetimes.datetime") as mock_datetime:
        mock_datetime.now.return_value = datetime(2022, 1, 1, tzinfo=timezone.utc)
        result = now_utc(naive=False)
        assert result == datetime(2022, 1, 1, tzinfo=timezone.utc)


def test_now_utc_naive():
    with patch("zsuite.helpers.datetimes.datetime") as mock_datetime:
        mock_datetime.now.return_value = datetime(2022, 1, 1, tzinfo=timezone.utc)
        result = now_utc(naive=True)
        assert result == datetime(2022, 1, 1).replace(tzinfo=None)


def test_now_utc():
    ts = now_utc()
    assert isinstance(ts, datetime)
    ts_naive = now_utc(naive=False)
    assert isinstance(ts_naive, datetime)


def test_normalize_from_date():
    input_date = date(2024, 2, 26)
    expected_datetime = datetime(2024, 2, 26, tzinfo=ZoneInfo("UTC"))
    assert normalize_timestamp(input_date) == expected_datetime


def test_normalize_from_naive_datetime():
    input_datetime = datetime(2024, 2, 26, 15, 30)
    expected_datetime = datetime(2024, 2, 26, 15, 30, tzinfo=ZoneInfo("UTC"))
    assert normalize_timestamp(input_datetime) == expected_datetime


def test_normalize_from_aware_datetime():
    input_datetime = datetime(2024, 2, 26, 15, 30, tzinfo=ZoneInfo("Europe/Paris"))
    expected_datetime = input_datetime.astimezone(ZoneInfo("UTC"))
    assert normalize_timestamp(input_datetime) == expected_datetime


def test_normalize_with_different_timezone():
    input_date = date(2024, 2, 26)
    expected_datetime = datetime(2024, 2, 26, tzinfo=ZoneInfo("America/New_York"))
    assert normalize_timestamp(input_date, tz="America/New_York") == expected_datetime


def test_normalize_raises_type_error():
    with pytest.raises(TypeError):
        normalize_timestamp("2024-02-26")  # passing a string instead of a date/datetime



from datetime import datetime, timedelta
from random import uniform
from time import sleep, time


def exponential_delay(
    minimum_sleep: int | float = 0.1,
    max_sleep: int | float = 60,
    jitter_pct: int | float = 0.1,
    backoff_factor: float = 1.25,
    cutoff: datetime | timedelta | int | None = None,
    max_attempts: int | None = None,
    enable_sleep: bool = True,
):
    """
    Generates sleep times based on exponential backoff algorithm.

    NOTE: First sleep is always 0 seconds. This is to allow the caller to perform an initial action before sleeping.

    Parameters:
    - minimum_sleep (float): The initial sleep time in seconds. Must be greater than 0.
    - max_sleep (float): The maximum sleep time in seconds. Must be greater than 0.
    - backoff_factor (float): The multiplier applied to the current sleep time to get the next sleep time.
    - jitter_pct (int | float): Optional jitter percentage to apply.
        - If int, should be between 0 and 100.
        - If float, should be between 0 and 1.
    - cutoff (datetime | timedelta | int | None): Optional cutoff time for the sleep.
        - If datetime, uses the given datetime as the cutoff.
        - If timedelta, adds the delta to the current time to get the cutoff.
        - If int, adds the number of seconds to the current time to get the cutoff.
    - enable_sleep (bool): Whether to actually sleep for the generated sleep times. Useful for testing.

    Yields:
    - sleep_time (float): The next sleep time in seconds.

    Raises:
    - ValueError: If any of the parameters are out of their valid range.

    Examples::

    for sleep_time in exponential_delay(initial_sleep=0.1, max_sleep=10, backoff_factor=1.15):
        print(f"Slept for {sleep_time} seconds")


    """
    current_delay = minimum_sleep
    count = 1

    # Validate & normalize arguments
    _validate_arguments(backoff_factor, max_attempts, max_sleep, minimum_sleep)
    cutoff = _normalize_cutoff(cutoff)
    jitter_pct = _normalize_jitter_pct(jitter_pct)

    # Main loop

    yield 0  # Always yield at least once, and the first sleep should be 0
    while True:
        count += 1
        if max_attempts is not None and count > max_attempts:
            return
        if cutoff and time() >= cutoff:
            return  # stop iteration if we've reached the cutoff time

        this_sleep = current_delay  # only apply jitter to the current sleep
        if jitter_pct:
            this_sleep = _apply_jitter(this_sleep, jitter_pct, minimum_sleep, max_sleep)
        this_sleep = _apply_bounds(this_sleep, minimum_sleep, max_sleep)

        if cutoff and time() + this_sleep > cutoff:
            this_sleep = cutoff - time()

        # Sleep (assuming it isn't disabled) and yield the sleep time
        if enable_sleep:
            sleep(this_sleep)  # sleep before yielding the slept time
        yield this_sleep

        if current_delay < max_sleep:
            current_delay *= backoff_factor
        current_delay = min(current_delay, max_sleep)


def _normalize_jitter_pct(jitter_pct):
    match jitter_pct:
        case int():
            if jitter_pct < 0 or jitter_pct > 100:
                raise ValueError("Jitter must be  int between 0 and 100 or float between 0 and 1")
            jitter_pct = float(jitter_pct * 0.01)
        case float():
            if jitter_pct < 0 or jitter_pct > 1:
                raise ValueError("Jitter must be float between 0 and 1, or int between 0 and 100")
    return jitter_pct


def _normalize_cutoff(cutoff):
    match cutoff:
        case datetime():
            cutoff = cutoff.timestamp()
        case timedelta():
            cutoff = time() + cutoff.total_seconds()
        case int():
            cutoff = cutoff + time()
    return cutoff


def _validate_arguments(backoff_factor, max_attempts, max_sleep, minimum_sleep):
    if minimum_sleep <= 0:
        raise ValueError("Initial sleep time must be greater than 0")
    if max_sleep <= 0:
        raise ValueError("Maximum sleep time must be greater than 0")
    if backoff_factor <= 1:
        raise ValueError("Backoff factor must be greater than 1")
    if max_attempts is not None and max_attempts <= 0:
        raise ValueError("If provided, max_attempts must be greater than 0")


def _apply_bounds(this_sleep, minimum_sleep, max_sleep):
    if this_sleep < minimum_sleep:
        this_sleep = minimum_sleep
    if this_sleep > max_sleep:
        this_sleep = max_sleep
    return this_sleep


def _apply_jitter(this_sleep, jitter_pct, minimum_sleep, max_sleep):
    upper = this_sleep + this_sleep * jitter_pct
    lower = this_sleep - this_sleep * jitter_pct
    if lower < minimum_sleep:
        lower = minimum_sleep
    if upper > max_sleep:
        upper = max_sleep
    this_sleep += uniform(
        lower,
        upper,
    )
    return this_sleep



import os
from pathlib import Path
from unittest.mock import mock_open, patch

import pytest
from cryptography.fernet import Fernet

from zsuite.config import load_config, load_env
from zsuite.exceptions import MissingVaultKey


def test_load_config():
    # Generate a key for encryption
    key = Fernet.generate_key()
    cipher_suite = Fernet(key)
    encrypted_password = cipher_suite.encrypt(b"test_password")

    # Mock configuration data
    config_data = f"""
    database:
        host: localhost
        password: !secret {encrypted_password.decode()}
    """
    # Mock environment variables and file opening
    with (
        patch.dict(os.environ, {"CONFIG_FILE": "/fake/path", "VAULT_KEY": key.decode()}),
        patch("builtins.open", mock_open(read_data=config_data)),
    ):
        config = load_config()
        assert config == {"database": {"host": "localhost", "password": "test_password"}}


def test_load_config_no_config_file_env_var():
    with patch.dict(os.environ, {}, clear=True), pytest.raises(Exception, match="CONFIG_FILE not set"):
        load_config()


def test_load_config_encrypted_value_no_decryption_key():
    # Mock configuration data with encrypted password
    config_data = """
    database:
        host: localhost
        password: !secret encryptedpassword
    """
    # Mock environment variables and file opening
    with (
        patch.dict(os.environ, {"CONFIG_FILE": "/fake/path"}),
        patch("builtins.open", mock_open(read_data=config_data)),
        pytest.raises(
            MissingVaultKey,
            match="Encrypted value found but no decryption key provided.",
        ),
    ):
        load_config()


def test_load_env():
    test_file_path = Path(__file__).parent.absolute()
    env_file_path = test_file_path / ".test_env"

    load_env(env_file_path, required=True)
    assert os.getenv("TEST_VARIABLE") == "test_value"

    # Test case: File does not exist and required=True
    with pytest.raises(FileNotFoundError):
        load_env(".does_not_exist", required=True)

    # Test case: File does not exist and required=False (should print warning)
    load_env(".does_not_exist", required=False)



import logging
import os

from dotenv import load_dotenv

from .crypto import decrypt_fernet_str
from .exceptions import EncryptedValueError, MissingVaultKey
from .yaml import load_and_decrypt_yaml

# Sentinel value to represent an unset default
UNSET_DEFAULT = object()


def config_var(name: str, default: any = UNSET_DEFAULT) -> any:
    """
    Get a config variable from the service object or environment, prefer environment

    """
    value = default

    if name in os.environ:
        value = os.environ[name]
    # else:
    #     config = getattr(SVCObj().svc, "config", None)
    #     if config is not None:
    #         if name in config:
    #             value = config[name]
    #         elif name.lower() in config:
    #             value = config[name.lower()]

    if value is UNSET_DEFAULT:
        raise ValueError(f"Config variable {name} not set and no default provided")

    if isinstance(value, bytes):
        value = _attempt_decode(name, value)

    if isinstance(value, str):
        value = _normalize_config_string(name, value)

    return value


def _normalize_config_string(name, value):
    if value.lower() == "true":
        value = True
    elif value.lower() == "false":
        value = False
    elif value.startswith("!secret"):
        value = extract_secret(name, value)
    return value


def _attempt_decode(name, value):
    try:
        value = value.decode()
    except Exception as e:
        logging.error(f"Error decoding config var {name}: {e}")
        raise e
    return value


def extract_secret(name, value):
    vault_key = os.getenv("VAULT_KEY")
    if vault_key is None:
        raise MissingVaultKey("Encrypted Config encountered with no VAULT_KEY environment variable set")
    else:
        try:
            value = _decrypt_cfg_var(os.getenv("VAULT_KEY"), value)
        except Exception as e:
            logging.error(f"Error decrypting config var {name}: {e}")
            raise EncryptedValueError(f"Error decrypting config var {name}: {e}") from e
    return value


def _decrypt_cfg_var(key, value):
    value = value.removeprefix("!secret ")
    return decrypt_fernet_str(key, value)


def load_config(decryption_key=None, config_file=None) -> dict:
    """
    Loads a YAML configuration file, decrypting values tagged with "!secret" using the provided decryption key.

    The function registers a custom constructor to handle decryption of values tagged with "!secret" in the YAML file.
    If the decryption_key is not provided, the function uses the secret_constructor without the key.
    If the config_file is not provided, the function looks for the file path in the "CONFIG" environment variable.

    :param decryption_key: Optional; a string representing the Fernet decryption key to be used for decryption.
                           If not provided, the constructor will attempt to use the environment variable "VAULT_KEY".
    :param config_file: Optional; a string representing the path to the YAML configuration file to be loaded.
                        If not provided, the function will look for the path in the "CONFIG" environment variable.
    :return: A dictionary representing the contents of the loaded YAML file, with encrypted values decrypted
    :raises Exception: If neither the config_file parameter nor the "CONFIG" environment variable is set.
    """
    if decryption_key is None:
        decryption_key = os.getenv("VAULT_KEY", None)
    if config_file is None:
        config_file = os.getenv("CONFIG_FILE")
    if config_file is None:
        raise Exception("CONFIG_FILE not set")

    return load_and_decrypt_yaml(decryption_key, config_file)


def load_env(env_file=".env", required=False):
    if os.path.exists(env_file):
        load_dotenv(env_file)
    else:
        if required:
            raise FileNotFoundError(f"Environment file {env_file} not found")
        else:
            print(f"Warning: environment file {env_file} not found")



class ZSuiteException(Exception):
    """Base exception for zsuite."""


class UndeterminedBool(ZSuiteException):
    """Raised when a boolean value cannot be determined."""


class EncryptedValueError(ZSuiteException):
    """Raised when an encrypted value cannot be decrypted."""


class MissingVaultKey(ZSuiteException):
    """Raised when a VAULT_KEY environment variable is missing."""


class CircuitBreakerTripped(ZSuiteException):
    """Raised when a circuit breaker trips."""



from cryptography.fernet import Fernet

from zsuite.crypto import decrypt_fernet_str, encrypt_fernet_str
from zsuite.helpers import want_bytes


def test_round_trip_str():
    key = Fernet.generate_key().decode()
    value = "some_secret_data"
    encrypted_value = encrypt_fernet_str(key, value)
    decrypted_value = decrypt_fernet_str(key, encrypted_value)
    assert decrypted_value == value


def test_round_trip_bytes():
    key = Fernet.generate_key()
    value = b"some_secret_data"
    encrypted_value = encrypt_fernet_str(key, value)
    decrypted_value = decrypt_fernet_str(key, encrypted_value)
    assert decrypted_value == value.decode()


def test_mixed_types():
    key_str = Fernet.generate_key().decode()
    key_bytes = want_bytes(key_str)
    value_str = "some_secret_data"
    value_bytes = want_bytes(value_str)

    encrypted_value_str = encrypt_fernet_str(key_str, value_str)
    decrypted_value_str = decrypt_fernet_str(key_str, encrypted_value_str)
    assert decrypted_value_str == value_str

    encrypted_value_bytes = encrypt_fernet_str(key_bytes, value_bytes)
    decrypted_value_bytes = decrypt_fernet_str(key_bytes, encrypted_value_bytes)
    assert decrypted_value_bytes == value_bytes.decode()



from .backoff import exponential_delay
from .circuit_breaker import CircuitBreaker
from .config import config_var, load_config, load_env
from .file_operations import debug_file_path
from .logging import setup_logging



import logging
import os
import time

from .config import config_var

VALID_LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
DEFAULT_SUPPRESS_PACKAGES = [
    "urllib3",
    "requests",
    "pymongo",
    "apscheduler",
    "selenium",
    "werkzeug",
]
DEFAULT_LOG_FORMAT = "%(asctime)s %(levelname)-8s %(name)-20s %(threadName)-10s %(funcName)s : %(message)s"


def setup_logging(
    log_level: str | None = None,
    log_format: str | None = None,
    suppress_packages: list[str] | None = None,
    suppress_level: str = "WARNING",
    force_utc: bool = True,
    dateformat: str | None = None,
) -> None:
    log_level = log_level or config_var("LOG_LEVEL", "INFO")
    dateformat = dateformat or ("%Y-%m-%d %H:%M:%SZ" if force_utc else "%Y-%m-%d %H:%M:%S")
    logging.Formatter.converter = time.gmtime if force_utc else time.localtime

    log_format = log_format or config_var("LOG_FORMAT", DEFAULT_LOG_FORMAT)

    if log_level not in VALID_LOG_LEVELS:
        raise ValueError(f"Invalid log level: {log_level}")

    root_logger = logging.getLogger()

    # If TESTING is set, skip clearing handlers (for pytest/caplog compatibility)
    if os.environ.get("TESTING") == "true":
        logging.info("TESTING is set")
    else:
        # Resetting logging configuration to a known state
        root_logger.handlers.clear()

    # Configure new logging handler
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter(log_format, datefmt=dateformat))
    root_logger.addHandler(handler)
    root_logger.setLevel("INFO")  # Ensure setup_logging always logs setup
    logging.info("Logging level set to %s", log_level)
    root_logger.setLevel(log_level)

    supress_package_logs(suppress_level, suppress_packages)


def supress_package_logs(suppress_level, suppress_packages):
    # Suppress specific package logging
    suppress_packages = suppress_packages or DEFAULT_SUPPRESS_PACKAGES
    for pkg in suppress_packages:
        logging.getLogger(pkg).setLevel(suppress_level)



import time
from collections import deque

from .exceptions import CircuitBreakerTripped


class CircuitBreaker:
    def __init__(self, max_events: int = None, time_window: int = None, custom_exception=None):
        if max_events is None:
            raise ValueError("max_events must be set")
        if time_window is None:
            raise ValueError("time_window must be set")
        self.max_events = max_events
        self.time_window = time_window
        self._exception_type = custom_exception if custom_exception else CircuitBreakerTripped
        self._event_timestamps = deque()

    def increment(self, current_exception=None):
        if self.count() >= self.max_events:
            if current_exception:
                raise CircuitBreakerTripped(
                    f"Circuit Breaker tripped at max events: {current_exception}"
                ) from current_exception
            else:
                raise self._exception_type("Circuit Breaker tripped at max events") from None

        else:
            self._event_timestamps.append(time.time())

    def reset(self):
        self._event_timestamps.clear()

    def count(self):
        self._remove_old_events()
        return len(self._event_timestamps)

    def _remove_old_events(self):
        current_time = time.time()
        while self._event_timestamps and current_time - self._event_timestamps[0] > self.time_window:
            self._event_timestamps.popleft()



import logging
import os

from .config import config_var


def remove_if_exists(fpath: str):
    """
    Given a full path to a file, remove it if it exists.
    Returns none regardless of file's presence

    :param fpath: full path to file
    :return: None
    """
    try:
        os.remove(fpath)
        return True
    except OSError:
        return False


def debug_file_path(fpath: str, force_print: bool = False):
    # Determine if debugging should be enabled for logging
    should_debug = config_var("DEBUG_FILE_PATH", False) or config_var("LOG_LEVEL", "INFO") == "DEBUG"

    # Collect debug information
    debug_info = []
    summary_dict = {}

    debug_info.append(f"\n---------------------------\nFILE DEBUG for: {fpath}")
    summary_dict["file_path"] = fpath

    is_absolute = os.path.isabs(fpath)
    debug_info.append(f"Path is absolute: {'Yes' if is_absolute else 'No'}")
    summary_dict["is_absolute"] = is_absolute

    absolute_path = os.path.abspath(fpath)
    debug_info.append(f"Absolute path: {absolute_path}")
    summary_dict["absolute_path"] = absolute_path

    file_exists = os.path.isfile(fpath)
    debug_info.append(f"File exists: {'Yes' if file_exists else 'No'}")
    summary_dict["file_exists"] = file_exists

    file_readable = os.access(fpath, os.R_OK)
    debug_info.append(f"File is readable: {'Yes' if file_readable else 'No'}")
    summary_dict["file_readable"] = file_readable

    file_writable = os.access(fpath, os.W_OK)
    debug_info.append(f"File is writable: {'Yes' if file_writable else 'No'}")
    summary_dict["file_writable"] = file_writable

    file_executable = os.access(fpath, os.X_OK)
    debug_info.append(f"File is executable: {'Yes' if file_executable else 'No'}")
    summary_dict["file_executable"] = file_executable

    file_is_directory = os.path.isdir(fpath)
    debug_info.append(f"File is a directory: {'Yes' if file_is_directory else 'No'}")
    summary_dict["file_is_directory"] = file_is_directory

    parent_dir = os.path.dirname(absolute_path)
    parent_dir_exists = os.path.isdir(parent_dir)
    debug_info.append(f"Parent directory exists: {'Yes' if parent_dir_exists else 'No'}")
    summary_dict["parent_dir_exists"] = parent_dir_exists

    parent_dir_writable = os.access(parent_dir, os.W_OK)
    debug_info.append(f"Parent directory is writable: {'Yes' if parent_dir_writable else 'No'}")
    summary_dict["parent_dir_writable"] = parent_dir_writable

    parent_dir_readable = os.access(parent_dir, os.R_OK)
    debug_info.append(f"Parent directory is readable: {'Yes' if parent_dir_readable else 'No'}")
    summary_dict["parent_dir_readable"] = parent_dir_readable

    # Combine all debug messages into a multi-line string
    combined_debug_message = "\n".join(debug_info)
    combined_debug_message += "\n---------------------------\n"

    # Log at the appropriate level based on configuration, if should_debug is True
    if should_debug:
        log_level = config_var("LOG_LEVEL", "INFO")
        getattr(logging, log_level.lower())(f"{combined_debug_message}")

    if force_print:
        print(combined_debug_message)

    return summary_dict



import os

import yaml
from cryptography.fernet import Fernet

from zsuite.helpers import want_bytes

from .exceptions import MissingVaultKey


def secret_constructor(decryption_key: str | bytes = None):
    """Creates and returns a custom YAML constructor function for decrypting values tagged with "!secret".

    This function returns a constructor that can be used with a YAML loader to decrypt scalar values tagged with
    "!secret" in a YAML file. If a decryption_key is provided, it will be used to decrypt the values. If not,
    the function will attempt to use the "VAULT_KEY" environment variable as the decryption key.

    :param decryption_key: Optional; the Fernet decryption key to be used for decryption.
                           If not provided, the returned constructor will use the "VAULT_KEY" environment variable.
    :return: A constructor function that takes a loader and a node, decrypts the value if it's tagged with "!secret",
             and returns the decrypted or original value.
    :raises ValueError: If a value tagged with "!secret" is encountered but no decryption key is provided or found.
    """
    decryption_key = want_bytes(decryption_key)

    def _constructor(loader, node):
        value = loader.construct_scalar(node)
        key = decryption_key if decryption_key is not None else os.getenv("VAULT_KEY")
        if key is None:
            raise ValueError("Encountered encrypted value, but no decryption key was provided.")
        f = Fernet(key)
        return f.decrypt(value.encode()).decode()

    return _constructor


def load_and_decrypt_yaml(decryption_key=None, file_path=None):
    """Loads and parses a YAML file, decrypting values tagged with "!secret" using the provided decryption key.

    :param decryption_key: Optional; String representing the Fernet decryption key to be used for decryption.
    :param file_path: String representing the path to the YAML file to be loaded.
    :return: A dictionary representing the contents of the loaded YAML file, with encrypted values decrypted.
    """
    try:
        # First, read the file without applying any constructors to check for encrypted values
        with open(file_path) as f:
            file_content = f.read()
            if "!secret" in file_content:
                if decryption_key is None:
                    if "VAULT_KEY" in os.environ:
                        decryption_key = os.environ["VAULT_KEY"]
                    else:
                        raise MissingVaultKey("Encrypted value found but no decryption key provided.")
                yaml.constructor.SafeConstructor.add_constructor("!secret", secret_constructor(decryption_key))
        with open(file_path) as f:
            return yaml.safe_load(f)

    except FileNotFoundError:
        raise FileNotFoundError(f"File {file_path} not found") from None
    except NotADirectoryError:
        raise FileNotFoundError(f"File {file_path} directory not found") from None
    except MissingVaultKey:
        raise
    except Exception as e:
        raise Exception(f"Error loading YAML file {file_path}: {e}") from e



from cryptography.fernet import Fernet

from .helpers import want_bytes


def decrypt_fernet_str(key: str | bytes, value: str | bytes) -> str:
    f = Fernet(want_bytes(key))
    return f.decrypt(want_bytes(value)).decode()


def encrypt_fernet_str(key: str | bytes, value: str | bytes) -> str:
    f = Fernet(want_bytes(key))
    return f.encrypt(want_bytes(value)).decode()



import logging
from unittest.mock import patch

import pytest

from zsuite import setup_logging
from zsuite.logging import DEFAULT_SUPPRESS_PACKAGES


def test_setup_logging_default(monkeypatch):
    monkeypatch.setenv("LOG_LEVEL", "INFO")
    monkeypatch.setenv(
        "LOG_FORMAT",
        "%(asctime)s %(levelname)-8s %(name)-20s %(threadName)-10s : %(message)s",
    )

    with (
        patch("logging.StreamHandler") as mock_handler,
        patch("logging.Formatter") as mock_formatter,
        patch("logging.getLogger") as mock_get_logger,
        patch("logging.info") as mock_info,
    ):
        setup_logging()

        mock_handler.assert_called()
        mock_formatter.assert_called_with(
            "%(asctime)s %(levelname)-8s %(name)-20s %(threadName)-10s : %(message)s",
            datefmt="%Y-%m-%d %H:%M:%SZ",
        )
        mock_get_logger.assert_called()
        mock_info.assert_called_with("Logging level set to %s", "INFO")


def test_setup_logging_custom(monkeypatch):
    monkeypatch.setenv("LOG_LEVEL", "DEBUG")
    monkeypatch.setenv("LOG_FORMAT", "custom_format")

    with (
        patch("logging.StreamHandler") as mock_handler,
        patch("logging.Formatter") as mock_formatter,
        patch("logging.getLogger") as mock_get_logger,
        patch("logging.info") as mock_info,
    ):
        setup_logging(log_level="DEBUG", log_format="custom_format")

        mock_get_logger.assert_called()
        mock_info.assert_called_with("Logging level set to %s", "DEBUG")

        mock_handler.assert_called()
        mock_formatter.assert_called_with("custom_format", datefmt="%Y-%m-%d %H:%M:%SZ")


def test_default_logging_setup(caplog, set_testing):
    """Test the default logging setup."""
    with caplog.at_level(logging.INFO):
        setup_logging()
        assert "Logging level set to INFO" in caplog.text
        assert caplog.records[0].levelname == "INFO"


def test_custom_log_level(caplog, set_testing):
    """Test logging setup with a custom log level."""
    custom_level = "DEBUG"
    with caplog.at_level(logging.DEBUG):
        setup_logging(log_level=custom_level)
        assert f"Logging level set to {custom_level}" in caplog.text
        assert logging.getLogger().getEffectiveLevel() == logging.DEBUG


def test_invalid_log_level():
    """Test logging setup with an invalid log level."""
    with pytest.raises(ValueError):
        setup_logging(log_level="INVALID")


def test_suppress_packages(caplog):
    """Test that package logging is suppressed correctly."""
    setup_logging(suppress_level="ERROR")
    for package in DEFAULT_SUPPRESS_PACKAGES:
        assert logging.getLogger(package).getEffectiveLevel() == logging.ERROR


def test_custom_suppress_packages(caplog):
    """Test custom packages are suppressed."""
    custom_packages = ["custom_package"]
    setup_logging(suppress_packages=custom_packages, suppress_level="CRITICAL")
    for package in custom_packages:
        assert logging.getLogger(package).getEffectiveLevel() == logging.CRITICAL



import os

import pytest
from cryptography.fernet import Fernet

from zsuite.config import config_var
from zsuite.exceptions import EncryptedValueError, MissingVaultKey


# Test case for retrieving a config variable from os.environ
def test_config_var_from_os_environ(monkeypatch):
    monkeypatch.setenv("MY_CONFIG_VAR", "my_value")
    result = config_var("MY_CONFIG_VAR")
    assert result == "my_value"


# Test case for using default value when config variable is not found
def test_config_var_with_default():
    result = config_var("NON_EXISTING_VAR", default="default_value")
    assert result == "default_value"
    assert config_var("SOME_OTHER_VAR1", default=True) is True
    assert config_var("SOME_OTHER_VAR2", default=False) is False
    assert config_var("SOME_OTHER_VAR2", default=None) is None


# Test case for handling unset config variable and no default provided
def test_config_var_unset_and_no_default():
    with pytest.raises(
        ValueError,
        match="Config variable NON_EXISTING_VAR not set and no default provided",
    ):
        config_var("NON_EXISTING_VAR")


def test_config_var_boolean():
    trues = ["true", "True", "TRUE"]
    falses = ["false", "False", "FALSE"]
    for t in trues:
        os.environ["TEST_VAR"] = t
        assert config_var("TEST_VAR") is True
    for f in falses:
        os.environ["TEST_VAR"] = f
        assert config_var("TEST_VAR") is False


def test_config_var_secret(monkeypatch):
    test_pw = "test_password123!"
    key = Fernet.generate_key()
    cipher_suite = Fernet(key)
    encrypted_password = cipher_suite.encrypt(test_pw.encode())
    env_str = f"!secret {encrypted_password.decode()}"
    monkeypatch.setenv("SECRET_CFG_VAR", env_str)

    # Test case: secret_constructor without key
    with pytest.raises(MissingVaultKey):
        result = config_var("SECRET_CFG_VAR")

    # Test case: secret_constructor with invalid key
    monkeypatch.setenv("VAULT_KEY", "invalid_key")
    with pytest.raises(EncryptedValueError):
        result = config_var("SECRET_CFG_VAR")

    # Test case: secret_constructor with key
    monkeypatch.setenv("VAULT_KEY", key.decode())
    result = config_var("SECRET_CFG_VAR")
    assert result == test_pw



from datetime import datetime, timedelta
from time import time

import pytest

from zsuite import exponential_delay
from zsuite.backoff import (
    _apply_bounds,
    _normalize_cutoff,
    _normalize_jitter_pct,
    _validate_arguments,
)


def test_initial_conditions():
    with pytest.raises(ValueError):
        list(exponential_delay(minimum_sleep=0, enable_sleep=False))

    with pytest.raises(ValueError):
        list(exponential_delay(max_sleep=0, enable_sleep=False))

    with pytest.raises(ValueError):
        list(exponential_delay(backoff_factor=1, enable_sleep=False))


def test_jitter_conditions():
    with pytest.raises(ValueError):
        list(exponential_delay(jitter_pct=-1, enable_sleep=False))

    with pytest.raises(ValueError):
        list(exponential_delay(jitter_pct=101, enable_sleep=False))


def test_cutoff():
    total_delay = 1
    max_sleep = 0.25
    cutoff = datetime.now() + timedelta(seconds=total_delay)
    delays = list(exponential_delay(minimum_sleep=0.001, max_sleep=max_sleep, cutoff=cutoff))
    assert sum(delays) <= total_delay + max_sleep + max_sleep * 0.1


def test_max_attempts():
    delays = list(exponential_delay(minimum_sleep=1, max_sleep=10, max_attempts=3, enable_sleep=False))
    assert len(delays) == 3  # Should have 3 elements including the first 0


def test_minimum_sleep():
    delays = list(exponential_delay(minimum_sleep=1, jitter_pct=0.9, max_attempts=100, enable_sleep=False))
    assert delays[0] == 0  # The first sleep should be 0
    assert all(d >= 1 for d in delays[1:])  # All other sleeps should be >= 1


def test_maximum_sleep():
    delays = list(
        exponential_delay(
            minimum_sleep=1,
            max_sleep=10,
            jitter_pct=0.5,
            max_attempts=100,
            enable_sleep=False,
        )
    )
    assert delays[0] == 0  # The first sleep should be 0
    assert all(1 <= d <= 10 for d in delays[1:])  # All other sleeps should be >= 1


def test_first_sleep():
    first_sleep = next(exponential_delay(minimum_sleep=1, max_sleep=10))
    assert first_sleep == 0  # The first sleep should be 0


def test_backoff():
    max_attempts = 15
    delays = list(
        exponential_delay(
            minimum_sleep=0.1,
            jitter_pct=0.0,
            max_sleep=60,
            max_attempts=max_attempts,
            enable_sleep=False,
            backoff_factor=2,
        )
    )
    expected = [
        0,
        0.1,
        0.2,
        0.4,
        0.8,
        1.6,
        3.2,
        6.4,
        12.8,
        25.6,
        51.2,
        60,
        60,
        60,
        60,
    ]
    assert len(delays) == max_attempts
    assert delays == expected


# Tests for _normalize_jitter_pct function
def test_normalize_jitter_pct_with_int():
    assert _normalize_jitter_pct(10) == 0.1
    assert _normalize_jitter_pct(100) == 1.0
    assert _normalize_jitter_pct(0) == 0.0


def test_normalize_jitter_pct_with_float():
    assert _normalize_jitter_pct(0.1) == 0.1
    assert _normalize_jitter_pct(1.0) == 1.0
    assert _normalize_jitter_pct(0.0) == 0.0


def test_normalize_jitter_pct_with_invalid_int():
    with pytest.raises(ValueError):
        _normalize_jitter_pct(-1)
    with pytest.raises(ValueError):
        _normalize_jitter_pct(101)


def test_normalize_jitter_pct_with_invalid_float():
    with pytest.raises(ValueError):
        _normalize_jitter_pct(-0.1)
    with pytest.raises(ValueError):
        _normalize_jitter_pct(1.1)


# Tests for _normalize_cutoff function
def test_normalize_cutoff_with_datetime():
    dt = datetime.now()
    assert round(_normalize_cutoff(dt) - dt.timestamp(), 3) == 0.0


def test_normalize_cutoff_with_timedelta():
    td = timedelta(seconds=10)
    assert round(_normalize_cutoff(td) - (time() + td.total_seconds()), 3) == 0.0


def test_normalize_cutoff_with_int():
    assert round(_normalize_cutoff(10) - (time() + 10), 3) == 0.0


# Tests for _validate_arguments function
def test_validate_arguments_all_good():
    _validate_arguments(1.1, 1, 1, 0.1)  # Should not raise any exception


def test_validate_arguments_bad_minimum_sleep():
    with pytest.raises(ValueError, match="Initial sleep time must be greater than 0"):
        _validate_arguments(1.1, 1, 1, 0)


def test_validate_arguments_bad_max_sleep():
    with pytest.raises(ValueError, match="Maximum sleep time must be greater than 0"):
        _validate_arguments(1.1, 1, 0, 0.1)


def test_validate_arguments_bad_backoff_factor():
    with pytest.raises(ValueError, match="Backoff factor must be greater than 1"):
        _validate_arguments(1, 1, 1, 0.1)


def test_validate_arguments_bad_max_attempts():
    with pytest.raises(ValueError, match="If provided, max_attempts must be greater than 0"):
        _validate_arguments(1.1, 0, 1, 0.1)


# Tests for _apply_bounds function
def test_apply_bounds_within_range():
    assert _apply_bounds(5, 1, 10) == 5  # Inside the range, should return the value itself


def test_apply_bounds_equal_to_min():
    assert _apply_bounds(1, 1, 10) == 1  # Equal to minimum, should return the minimum


def test_apply_bounds_equal_to_max():
    assert _apply_bounds(10, 1, 10) == 10  # Equal to maximum, should return the maximum


def test_apply_bounds_below_min():
    assert _apply_bounds(0, 1, 10) == 1  # Below minimum, should return the minimum


def test_apply_bounds_above_max():
    assert _apply_bounds(11, 1, 10) == 10  # Above maximum, should return the maximum
